;CAML16F.S -- Forth Kernel
; CamelForth for the Zilog ZNEO (Z16F Series)
; portions (C) 2006 Douglas Beattie Jr.
;
; ..based on CamelForth for the Zilog Z80,
; (C) 1994 Bradford J. Rodriguez
; Permission is granted to freely copy, modify,
; and distribute this program for personal or
; educational use.  Commercial inquiries should
; be directed to the author at 115 First St.,
; #105, Collingwood, Ontario L9Y 4W3, Canada
;
; CAML16F.S: Code Primitives
;   Source code is for the ZDS-II ZNEO 4.10.2 macro
;   assembler (provided with Zilog Developer
;   Studio II).
;
;   Forth words are documented as follows:
;x   NAME     stack -- stack    description
;   where x=C for ANS Forth Core words, X for ANS
;   Extensions, Z for internal or private words.
;
; Direct-Threaded Forth model for Zilog ZNEO
; 32 bit cell, 8 bit char, 8 bit (byte) address unit
; with typically 16-bit alignment.
;
;
; -----  CPU Register Assignments:
;
TOS EQU R9   ;Top of Stack
WP  EQU R10  ;Working Register/Pointer
RSP EQU R11  ;Return Stack Pointer
UAP EQU R12  ;User Area Pointer
IP  EQU R13  ;Instruction Pointer (Interpreter Pointer)
PSP EQU R15  ;Parameter Stack Pointer (is hardware SP)
;
;   R0 through R8 are available for user code.
;   Note also that R14 is reserved, as it has a special
;   function for stack frames.
;
SIZE_TIB  EQU 128 ; Terminal Input Buffer size
SIZE_UA   EQU 256 ; User Area size
SIZE_RSTK EQU 256 ; Return Stack Pointer size
SIZE_HOLD EQU 80 ; Hold Area size
SIZE_PAD  EQU 88  ; Pad Buffer size
SIZE_PSTK EQU 256 ; Parameter Stack Pointer size
MEM_END   EQU %FFBF00 ; 256 below Linker FAR_BSS section

; ----- ZNEO Camel Forth RAM MEMORY MAP
;
; Memory map for Z16F2811 using internal RAM only
; 000200h  Forth kernel in bottom of 128K Flash
; FFB000h  Terminal Input Buffer, 128 bytes (ZNEO on-chip RAM)
; FFB080h  Forth dictionary (ZNEO on-chip RAM)
; .... (2776 bytes of) Available Dictionary Space
; EM-3A8h  User area, 256 bytes
; EM-2A8h  Return stack, 256 bytes, grows down
;       Note: the Leave stack grows upward and shares
;       space allocated for the Return Stack
; EM-1A8h  HOLD area, 80 bytes, grows down
; EM-158h  PAD buffer, 88 bytes
; EM-100h  Parameter stack, 256 bytes, grows down
; FFBF00h  End of Memory ('EM') (End of RAM + 1)
; See also the definitions of U0, S0, and R0
; in the "system variables & constants" area.
; A task w/o terminal input requires 936 (3A8h) bytes.
; Double all except TIB and PAD for 32-bit CPUs.
;

;
; Revision history:
;   19 Aug 94 v1.0  Z80 Version, Direct-Threaded
;   25 Jan 95 v1.01  now using BDOS function 0Ah
;       for interpreter input; TIB at 82h.
;   02 Mar 95 v1.02  changed ALIGN to ALIGNED in
;       S" (S"); changed ,BRANCH to ,XT in DO.
;   25 SEP 2001 PORTED TO eZ80 by D. Beattie Jr.
;   OCTOBER 2006 v1.00 for ZNEO, ZDS II (ZiLOG) Tools
;


    INCLUDE "zneo.inc"

    XDEF _main
    XREF _init_uart  ; :EROM
    XREF _SYS_CLK_FREQ   ; :EROM
    XREF _far_stack:ERAM

CPUFREQ EQU 18432000 ;_SYS_CLK_FREQ
BAUDRATE_U0 EQU 57600
;;;UART0_DIVISOR EQU (CPUFREQ / (BAUDRATE_U0 * 16))


    DEFINE KERNEL,SPACE=ROM

CELLSIZE EQU 4 ;32-BIT
WORDSIZE EQU 2 ;16-BIT ALIGNMENT FOR CODE

; ===============================================
; Macros to define Forth headers
; HEAD  label,name,action
; IMMED label,name,action
;    label  = assembler name for this word
;             (special characters not allowed)
;    name   = Forth's name for this word
;    action = code routine for this word, e.g.
;             DOCOLON, or DOCODE for code words
; IMMED defines a header for an IMMEDIATE word.
;
; NOTE: length of name field is now automatically
; calculated by the macro, and is no longer a parameter
;
docode  EQU 0      ; flag to indicate CODE words

; link to previous Forth word
dlink  .SET 0     ; dictionary link to previous Forth word

; ---------------------------------------------
head    MACRO   label,length,name,action
       .ALIGN CELLSIZE
        DL dlink
        DL 0
$$:
dlink    .SET $B
        DB length,name
$$:
label:
        IFSAME  action,docode
		ELSE
        call action
        ENDIF
        ENDMACRO head

; ---------------------------------------------
immed   MACRO   label,length,name,action
       .ALIGN CELLSIZE
        DL dlink
        DL 1
$$:
dlink    .SET $B
        DB length,name
$$:
label:
        IFSAME action,docode
		ELSE
        call action
        ENDIF
        ENDMACRO immed

; ---------------------------------------------
; This NEXT macro would assemble the 'next'
; code in-line in every CamelForth CODE word.

;next    MACRO
;    LD WP,(IP++)
;    JP (WP)
;    ENDMACRO next

; We'll use a global 'next' for test & debug.
next    MACRO
    JPA dbg_next
    ENDMACRO next

; ---------------------------------------------


	SEGMENT KERNEL


; RESET AND INTERRUPT VECTORS ===================
;
;	Reset vector
;
 XDEF _IDENT
    ORG 0200h-32
_IDENT:
    ASCII "ZILOG ZNEO CAMEL FORTH",0

    ORG 0200h ;; 800000h

; On entry to main() the kernel begins execution.
; initial stacks were set up, and the DATA section
; was copied to RAM.

_main:

ENTRY:

    LD Caller_SP,SP ;save Stack pointer to caller for BYE
    ;CALL _init_Ports

; ********************************************************
; ** ( initialize registers, and jump to COLD )

	;processor SP = PSP at MEM_END, grows downward
    LD WP,#MEM_END-SIZE_PSTK+11 ;
    LD SPOV,WP
    LD WP,#MEM_END
    LD PSP,WP  ;_far_stack:ERAM

	; RSP at EM-100h, grows downward
    LEA RSP,-(SIZE_HOLD+SIZE_PAD+SIZE_PSTK)(PSP)

	;UAP, ;User Area begin at EM-200h
    LEA UAP,-(SIZE_UA+SIZE_RSTK)(RSP)

	;Init IP to do restart if COLD returns
    LEA IP,COLD ;RESTARTER

    ; Stuff CR (carriage return) into Terminal Input Buffer
    LEA WP,-SIZE_TIB(WP)

    LD WP,(TIB_pointer)
    LD.B (WP++),#0Dh
    LD.B (WP++),#0Ah

    CALL ENABLE_EXT_CLOCK
    CALL INIT_UART0

    LD R1,#55h ;sanity check - test console UART
	CALL TXMCHR_U0 ;transmit 'U' to terminal


	JP COLD      ; enter top-level Forth word

RESTARTER:
    JP ENTRY ;to Reset

dbg_next:
    LD WP,(IP++)
    JP (WP)


DELAY EQU 10000
ENABLE_EXT_CLOCK:
	LD R1,#%E0      ; external clock, R1 = #%E0
    CALL WRITE_OSCCTL   ; Write to Oscillator Control
    LD R0,#10000    ; Delay value
    DJNZ R0,$       ; Wait for the clock to stabilize
	LD R1,#%E1      ; Set the system clock as external
    CALL WRITE_OSCCTL   ; Write to Oscillator Control
    RET

;Write the byte in R1 to Oscillator Control Register
;Unlock the register before writing
WRITE_OSCCTL:
	LD R0,#%E7  ;Unlock Sequence, 1st byte
	LD.B OSCCTL,R0
	LD R0,#%18  ;Unlock Sequence, 2nd byte
	LD.B OSCCTL,R0            ;   Unlock the OSCCTL register
	LD.B OSCCTL,R1            ;   WriteEnable/Disable the external clock
    RET

INIT_UART0:
    ;;LD R0,#UART0_DIVISOR

    ; Calculate Divisor: (CPUFREQ / (BAUDRATE_U0 * 16))
    LD R1,#BAUDRATE_U0
    LD R0,#16
    MUL R1,R0   ; Product, innermost parentheses
    LD R0,#CPUFREQ
    UDIV R0,R1    ; Quotient, outer paren.

    CLR.B U0CTL0  ;Disable Uart0
    
    ;Store Divisor into Baud-rate Generator
    LD.W U0BR,R0

    ;Set up I/O pins for Tx/Rx
    LD R0,#30h      ;Port A, bits 5,4
    LD.B PAAFL,R0   ;Enable Alternate Function

    ; Enable Transmitter and Receiver
    LD R0,#(UxCTL0.TEN | UxCTL0.REN)
    LD.B U0CTL0,R0  ;Write Uart0 Control Register
    RET

;
; transmit byte in register R1
;
TXMCHR_U0:
; ZNEO - Data byte to send is in R1
    LD   R0,#UxSTAT0.TDRE  ; bit mask, TXDREmpty
$$: TM.B U0STAT0,R0  ;test UART0 Status Register
    JP Z,$B ;loop until
    LD.B U0TXD,R1 ;Send the byte
    RET

TXMCHR_U1:
    LD   R0,#UxSTAT0.TDRE  ; bit mask, TXDREmpty
$$: TM.B U1STAT0,R0  ;test UART0 Status Register
    JP Z,$B ;loop until
    LD.B U1TXD,R1 ;Send the byte
	RET

;
; check if character ready to receive
; returns R1 status, non-zero=key waiting
;

RCXRDY_U0:
    LD   R0,#UxSTAT0.RDA
    TM.B U0STAT0,R0
    LD   R1,#%0
    JP   Z,$F
    LD   R1,#%1
$$: RET		;also sets NZ-flg if char waiting

RCXRDY_U1:
    LD   R0,#UxSTAT0.RDA
    TM.B U1STAT0,R0
    LD   R1,#%0
    JP   Z,$F
    LD   R1,#%1
$$: RET		;also sets NZ-flg if char waiting

;
; receive byte from UART
; returns char in R1
;
RCXCHR_U0:
    LD   R0,#UxSTAT0.RDA
$$: TM.B U0STAT0,R0
    JP   Z,$B
    LD.UB R1,U0RXD
	RET

RCXCHR_U1:
    LD   R0,#UxSTAT0.RDA
$$: TM.B U1STAT0,R0
    JP   Z,$B
    LD.UB R1,U1RXD
	RET

; INTERPRETER LOGIC =============================
; See also "defining words" at end of this file

;C EXIT     --      exit a colon definition
    head EXXIT,4,{"EXIT"},docode
        LD IP,(RSP++)  ; pop old IP from ret stk
        next

;Z lit      -- x    fetch inline literal to stack
; This is the primitive compiled by LITERAL.
    head LIT,3,{"lit"},docode
	    PUSH TOS        ; push old TOS
        LD TOS,(IP++)   ; fetch cell at IP to TOS,
        next        ;        advancing IP

;C EXECUTE   i*x xt -- j*x   execute Forth word
;C                           at 'xt'
    head EXECUTE,7,{"EXECUTE"},docode
    LD WP,TOS   ;W := TOS
    POP TOS     ; get new TOS
    JP (WP)     ;jump (W)


; DEFINING WORDS ================================

; ENTER, a.k.a. DOCOLON, entered by CALL ENTER
; to enter a new high-level thread (colon def'n.)
; (internal code fragment, not a Forth word)
; N.B.: DOCOLON must be defined before any
; appearance of 'docolon' in a 'word' macro!
docolon:               ; (alternate name)
enter:
    LD (--RSP),IP   ; push old IP on ret stack
    POP IP          ; IP <- param field address
    next


;C VARIABLE   --      define a Forth variable
;   CREATE 1 CELLS ALLOT ;
; Action of RAM variable is identical to CREATE,
; so we don't need a DOES> clause to change it.
    head VARIABLE,8,{"VARIABLE"},docolon
        DL CREATE,LIT,1,CELLS,ALLOT,EXXIT

; DOVAR, code action of VARIABLE, entered by CALL
; DOCREATE, code action of newly created words
docreate:
dovar:  ; -- a-addr
    POP WP          ; parameter field address
	PUSH TOS        ; push old TOS
    LD TOS,WP       ; pfa = variable's adrs -> TOS
    next

;C CONSTANT   n --      define a Forth constant
;   CREATE , DOES> (machine code fragment)
    head CONSTANT,8,{"CONSTANT"},docolon
        DL CREATE,COMMA,XDOES
; DOCON, code action of CONSTANT,
; entered by CALL DOCON
docon:  ; -- x
    POP WP          ; parameter field address
    PUSH TOS        ; push old TOS
    LD TOS,(WP++)   ; TOS <- contents of parameter field
    next

;Z USER     n --        define user variable 'n'
;   CREATE , DOES> (machine code fragment)
    head USER,4,{"USER"},docolon
        DL CREATE,COMMA,XDOES
; DOUSER, code action of USER,
; entered by CALL DOUSER
douser:  ; -- a-addr
    POP WP          ; parameter field address
    PUSH TOS        ; push old TOS
    LD TOS,(WP++)   ; TOS <- contents of parameter field
    ADD TOS,UAP     ; add offset to user base address
                    ; and put result in TOS
    next

; DODOES, code action of DOES> clause
; entered by       CALL fragment
;                  parameter field
;                       ...
;        fragment: CALL DODOES
;                  high-level thread
; Enters high-level thread with address of
; parameter field on top of stack.
; (internal code fragment, not a Forth word)
dodoes: ; -- a-addr
    LD (--RSP),IP   ; push old IP on ret stk
    POP IP          ; address of new thread -> IP
    POP WP          ; address of parameter field
    PUSH TOS        ; push old TOS onto stack
    LD TOS,WP       ; pfa -> new TOS
    next


;C EMIT     c --    output character to console

    head EMIT,4,{"EMIT"},docode
        LD R1,TOS
        CALL TXMCHR_U0
	    POP TOS
        next

;C KEY?     -- f    return true if char waiting
    head QUERYKEY,4,{"KEY?"},docode
	    PUSH TOS    ;PREPARE FOR NEW TOS
        CALL RCXRDY_U0
        LD TOS,R1       ; RESULT IN TOS
        next

;C KEY      -- c    get character from keyboard
    head KEY,3,{"KEY"},docode
	    PUSH TOS    ;PREPARE FOR NEW TOS
        CALL RCXCHR_U0
        LD TOS,R1       ; RESULT IN TOS
        next


;X BYE     i*x --    return to Host OS
    head BYE,3,{"bye"},docode
    ;next
    LD SP,Caller_SP ;save Stack pointer to caller for BYE
    RET


; STACK OPERATIONS ==============================

;C DUP      x -- x x      duplicate top of stack
    head DUP,3,{"DUP"},docode
pushtos:
	PUSH TOS
        next

;C ?DUP     x -- 0 | x x    DUP if nonzero
    head QDUP,4,{"?DUP"},docode
    OR TOS,TOS
    JP NZ,pushtos
    next

;C DROP     x --          drop top of stack
    head DROP,4,{"DROP"},docode
poptos:
	POP TOS
    next


;C SWAP     x1 x2 -- x2 x1    swap top two items
    head SWOP,4,{"SWAP"},docode
    POP WP
    PUSH TOS
    LD TOS,WP
    next

;C OVER    x1 x2 -- x1 x2 x1   per stack diagram
    head OVER,4,{"OVER"},docode
    LD WP,(PSP)
    PUSH TOS
    LD TOS,WP
    next

;C ROT    x1 x2 x3 -- x2 x3 x1  per stack diagram
    head ROT,3,{"ROT"},docode
    POP R0  ; x1 (x2) x3=TOS
    POP WP  ; (x1) x3=TOS
    PUSH R0 ; x2 x3=TOS
    PUSH TOS  ; x2 x3 x3=TOS
    LD TOS,WP ; x2 x3 x1=TOS
    next


;X NIP    x1 x2 -- x2           per stack diagram
    head NIP,3,{"NIP"},docode
    POP WP ;discard, keep previous TOS
   	next

;X TUCK   x1 x2 -- x2 x1 x2     per stack diagram
    head TUCK,4,{"TUCK"},docolon
        DL SWOP,OVER,EXXIT

;C >R    x --   R: -- x   push to return stack
    head TOR,2,{">R"},docode
    LD (--RSP),TOS ; push TOS onto rtn stk
	POP TOS    ; pop new TOS
    next

;C R>    -- x    R: x --   pop from return stack
    head RFROM,2,{"R>"},docode
    ; pop old IP from ret stk
	PUSH TOS        ; push old TOS
    LD TOS,(RSP++)  ; pop top rtn stk item
    next            ;to TOS

;C R@    -- x     R: x -- x   fetch from rtn stk
    head RFETCH,2,{"R@"},docode
	PUSH TOS       ; push old TOS
    LD TOS,(RSP) ; fetch top rtn stk item    
    next           ; to TOS

;Z SP@  -- a-addr       get data stack pointer
    head SPFETCH,3,{"SP@"},docode
	PUSH TOS
	LD TOS,PSP
    next

;Z SP!  a-addr --       set data stack pointer
    head SPSTORE,3,{"SP!"},docode
	;DI	;must disable INTs if you're using them!
    LD PSP,TOS
   	POP TOS     ; get new TOS
	;EI ;must re-enable INTs if you're using them!
    next

;Z RP@  -- a-addr       get return stack pointer
    head RPFETCH,3,{"RP@"},docode
	PUSH TOS
    LD TOS,RSP
    next

;Z RP!  a-addr --       set return stack pointer
    head RPSTORE,3,{"RP!"},docode
    LD RSP,TOS
	POP TOS
    next

; MEMORY AND I/O OPERATIONS =====================

;C !        x a-addr --   store cell in memory
    head STORE,1,{'!'},docode
    POP WP
    LD (TOS),WP
    POP TOS
    next

;C C!      char c-addr --    store char in memory
    head CSTORE,2,{"C!"},docode
    POP WP      ; W <-- char
    LD.B (TOS),WP	; ( a-addr ) <--- char
	POP TOS
    next

;C @       a-addr -- x   fetch cell from memory
    head FETCH,1,{'@'},docode
    LD WP,TOS
    LD TOS,(WP)
    next

;C C@     c-addr -- char   fetch char from memory
    head CFETCH,2,{"C@"},docode
    LD WP,TOS
    LD.B TOS,(WP)
    next


; ARITHMETIC AND LOGICAL OPERATIONS =============

;C +       n1/u1 n2/u2 -- n3/u3     add n1+n2
    head PLUS,1,{'+'},docode
    POP WP
    ADD TOS,WP
    next

;X M+       d n -- d         add single to double
    head MPLUS,2,{"M+"},docode
	
	;there is a 64-bit word at PSP, with n in TOS.
	; add 32-bit n to the 64-bit word
    LD  R0,TOS  ; move TOS, prepare for add
    POP WP      ; pop high 32 bits  (big-endian)
    POP TOS     ; pop low 32 bits  (big-endian)
    ADD TOS,R0  ; add n to low 32 bits
    ADC WP,#0   ; add any Carry bit
    PUSH TOS    ; low 32 bits back into PSP
    LD TOS,WP   ; and high 32 bits of sum is in TOS
    next

;C -      n1/u1 n2/u2 -- n3/u3    subtraction, n1-n2
;           minuend: n1,  subtrahend: n2
    head MINUS,1,{'-'},docode
    LD WP,TOS
    POP TOS
    SUB TOS,WP
    next

;C AND    x1 x2 -- x3            logical AND
    head zAND,3,{"AND"},docode
    POP WP
    AND TOS,WP
    next

;C OR     x1 x2 -- x3           logical OR
    head zOR,2,{"OR"},docode
    POP WP
    OR TOS,WP
    next

;C XOR    x1 x2 -- x3            logical XOR
    head zXOR,3,{"XOR"},docode
    POP WP
    XOR TOS,WP
    next

;C INVERT   x1 -- x2            bitwise inversion
    head INVERT,6,{"INVERT"},docode
    COM TOS
    next

;C NEGATE   x1 -- x2            two's complement
    head NEGATE,6,{"NEGATE"},docode
    NEG TOS
    next

;C 1+      n1/u1 -- n2/u2       add 1 to TOS
    head ONEPLUS,2,{"1+"},docode
	INC TOS
    next

;C 1-      n1/u1 -- n2/u2     subtract 1 from TOS
    head ONEMINUS,2,{"1-"},docode
	DEC TOS
    next

 COMMENT ~
;Z ><      x1 -- x2         swap bytes (not ANSI)
    head SWAPBYTES,2,{"><"},docode
;?? WANT TO SWAP ALL 4 BYTES, MAKE LITTLE-ENDIAN?
    next
 ~ENDCOMMENT

;C 2*      x1 -- x2         arithmetic left shift
    head TWOSTAR,2,{"2*"},docode
	ADD TOS,TOS
    next

;C 2/      x1 -- x2        arithmetic right shift
    head TWOSLASH,2,{"2/"},docode
    LD R0,#1
   	SRA TOS,R0
    next

;C LSHIFT  x1 u -- x2    logical L shift u places
    head LSHIFT,6,{"LSHIFT"},docode
    LD WP,TOS
    POP TOS
    SLL TOS,WP
    next

;C RSHIFT  x1 u -- x2    logical R shift u places
    head RSHIFT,6,{"RSHIFT"},docode
    LD WP,TOS
    POP TOS
    SRL TOS,WP
    next

;C +!     n/u a-addr --       add cell to memory
    head PLUSSTORE,2,{"+!"},docode
    POP WP
    LD R0,(TOS)
    ADD WP,R0
    LD (TOS),WP
    POP TOS
    next


; COMPARISON OPERATIONS =========================

;C 0=     n/u -- flag    return true if TOS=0
    head ZEROEQUAL,2,{"0="},docode
zero_equal:
	SUB TOS,#1  ; Set Carry Flag if TOS==0
    SBC TOS,TOS ; Difference is zero (False) if no Carry
    next        ; otherwise will be 0xFFFFFFFF

;C 0<     n -- flag      true if TOS negative
    head ZEROLESS,2,{"0<"},docode
	ADD TOS,TOS    ; sign bit -> Carry Flag
    SBC TOS,TOS    ; propagate that bit      
    next

;C =      x1 x2 -- flag         test x1=x2
    head zEQUAL,1,{'='},docode
    POP WP
    SUB TOS,WP  ;result will be zero if equal
    JP zero_equal
    ;;next

;X <>     x1 x2 -- flag    test not eq (not ANSI)
    head NOTEQUAL,2,{"<>"},docolon
        DL zEQUAL,ZEROEQUAL,EXXIT

;C <      n1 n2 -- flag        test n1<n2, signed
    head LESS,1,{'<'},docode
    POP WP
    CP WP,TOS
    LD TOS,#0 ;assume greater-or-equal
    JP GE,$F  ;skip next if so
    COM TOS   ; complement bits, result=TRUE
$$:
    next

;C >     n1 n2 -- flag         test n1>n2, signed
    head GREATER,1,{'>'},docolon
        DL SWOP,LESS,EXXIT

;C U<    u1 u2 -- flag       test u1<n2, unsigned
    head ULESS,2,{"U<"},docode
    POP WP
    SUB WP,TOS
    SBC TOS,TOS ; propagate Carry Flag
    next

;X U>    u1 u2 -- flag     u1>u2 unsgd (not ANSI)
    head UGREATER,2,{"U>"},docolon
        DL SWOP,ULESS,EXXIT


; LOOP AND BRANCH OPERATIONS ====================

;Z branch   --                  branch always
    head BRANCH,6,{"branch"},docode
dobranch:
    LD WP,(IP++)
    LD IP,WP           ; get inline value => IP
    next

;Z ?branch   x --              branch if TOS zero
    head QBRANCH,7,{"?branch"},docode
    OR TOS,TOS      ; test old TOS
    POP TOS         ; pop new TOS
    JP Z,dobranch   ;if old TOS=0, branch
    LEA IP,4(IP)    ; else skip inline value
    next



;Z (do)    n1|u1 n2|u2 --  R: -- sys1 sys2
;Z                          run-time code for DO
; '83 and ANSI standard loops terminate when the
; boundary of limit-1 and limit is crossed, in
; either direction.  This can be conveniently
; implemented by making the limit 8000h, so that
; arithmetic overflow logic can detect crossing.
; I learned this trick from Laxen & Perry F83.
; fudge factor = 8000h-limit, to be added to
; the start value.
    head XDO,4,{"(do)"},docode

    POP WP  ; limit into WP

    LD R0,#%80000000

    SUB R0,WP       ; 80000000h-limit in WP

    LD (--RSP),R0   ; push this fudge factor
                    ; onto return stack
                    ; for later use by 'I'

    ADD R0,TOS      ; add fudge to start value

    LD (--RSP),R0   ; push adjusted start value
                    ;    onto return stack
                    ;    as the loop index.

    POP TOS     ;pop new TOS
    next

;Z (loop)   R: sys1 sys2 --  | sys1 sys2
;Z                        run-time code for LOOP
; Add 1 to the loop index.  If loop terminates,
; clean up the return stack and skip the branch.
; Else take the inline branch.  Note that LOOP
; terminates when index=8000h.
    head XLOOP,6,{"(loop)"},docode
    LD R0,#1    ;default increment value = 1
looptst:
    LD WP,(RSP)
    ADD WP,R0       ; increment w/overflow test
    JP OV,loopterm  ; overflow=loop done
    LD (RSP),WP     ; save the updated index
    JP dobranch     ; take the inline branch
loopterm:
    ; terminate the loop -- drop 2 cells from RSP
    LEA RSP,8(RSP)  ; discard the loop info
    LEA IP,4(IP)    ; skip the inline branch
    next

;Z (+loop)   n --   R: sys1 sys2 --  | sys1 sys2
;Z                        run-time code for +LOOP
; Add n to the loop index.  If loop terminates,
; clean up the return stack and skip the branch.
; Else take the inline branch.
    head XPLUSLOOP,7,{"(+loop)"},docode
    LD R0,TOS   ; old TOS = loop increment
    POP TOS     ; new TOS
    JP looptst

;C I        -- n   R: sys1 sys2 -- sys1 sys2
;C                  get the innermost loop index
    head II,1,{'I'},docode
    PUSH TOS        ; push old TOS
    LD TOS,0(RSP)   ; get current loop index
    LD R0,4(RSP)    ; get fudge factor
    SUB TOS,R0  ; subtract fudge factor,
                ;  returning true index
    next


;C J        -- n   R: 4*sys -- 4*sys
;C                  get the second loop index
    head JJ,1,{'J'},docode

    PUSH TOS    ; push old TOS

    LD TOS,8(RSP)   ; get current loop index

    LD R0,12(RSP)   ; get fudge factor

    SUB TOS,R0      ; subtract fudge factor,
                    ; returning true index
    next



;C UNLOOP   --   R: sys1 sys2 --  drop loop parms
    head UNLOOP,6,{"UNLOOP"},docode

    LEA RSP,8(RSP)
    next


; MULTIPLY AND DIVIDE ===========================

;C UM*     u1 u2 -- ud   unsigned 32x32->64 mult.
    head UMSTAR,3,{"UM*"},docode
    POP WP
    UMUL TOS,WP
    PUSH TOS   ; lo result into stack
    LD TOS,WP  ; hi result in TOS
    next



;Z UM/MOD   ud u1 -- u2 u3   unsigned 64/32->32
;   divide unsigned double 'ud' by unsigned 'u1',
;   leaving remainder 'u2' with quotient 'u3' as TOS
    head UMSLASHMOD,6,{"UM/MOD"},docode
    POP R0  ; high-order quad, high 32 bits of ud
    POP R1  ; low-order quad, low 32 bits of ud

;WARNING:
    UDIV64 RR0,TOS
;   This could produce one of two different hardware
;   exceptions:  divide by zero, or divide overflow!
;   CPU Exception handlers should Jump to ABORT" with
;   a string describing the exception.
;
    PUSH R1     ;if all is well (no exception occured),
    LD TOS,R0   ;  push remainder and set new TOS
    next




; BLOCK AND STRING OPERATIONS ===================

;C FILL   c-addr u char --  fill memory with char
;       
    head FILL,4,{"FILL"},docode

    POP R0  ; count, # of bytes to fill
    POP WP
    OR R0,R0    ; sanity check -- count==0 ?
    JP Z,filldone
;;filloop:
$$: LD.B (WP++),TOS
    DEC R0
    JP NZ,$B
filldone:
    POP TOS
    next


;X CMOVE   c-addr1 c-addr2 u --  move from bottom
; as defined in the ANSI optional String word set
; On byte machines, CMOVE and CMOVE> are logical
; factors of MOVE.  They are easy to implement on
; CPUs which have a block-move instruction.
    head CMOVE,5,{"CMOVE"},docode

    ; count is in TOS
    OR TOS,TOS
    POP R1 ;destination of move
    POP R0 ;source of move
    JP Z,cmove_done

$$: LD.UB WP,(R0++)
    LD.B (R1++),WP
    DEC TOS
    JP NZ,$B

cmove_done:
    POP TOS
    next


;X CMOVE>  c-addr1 c-addr2 u --  move from top
; as defined in the ANSI optional String word set
    head CMOVEUP,6,{"CMOVE>"},docode
  ;NOT IMPLEMENTED...
  POP TOS
  POP TOS
  POP TOS
        next


;Z SKIP   c-addr u c -- c-addr' u'
;Z                          skip matching chars
; Although SKIP, SCAN, and S= are perhaps not the
; ideal factors of WORD and FIND, they closely
; follow the string operations available on many
; CPUs, and so are easy to implement and fast.
    head SKIP,4,{"SKIP"},docode
    LD R0,TOS  ; TOS contains char to skip e.g. skipwhite()
    POP TOS ; count, # of bytes to fill
    POP WP  ; current pointer to string
    OR TOS,TOS  ; sanity check -- count==0 ?
    JP Z,skip_done

$$: CP.B (WP),R0  ; looking for 1st mismatch...
    JP NZ,skip_done
    INC WP
    DEC TOS
    JP NZ,$B
skip_done:
    PUSH WP  ; push c-addr', u' already in TOS
    next




;Z SCAN    c-addr u c -- c-addr' u'
;Z                      find matching char
    head SCAN,4,{"SCAN"},docode
    LD R0,TOS  ; TOS contains char to match
    POP TOS ; count, # of bytes to fill
    POP WP  ; current pointer to string
    OR TOS,TOS  ; sanity check -- count==0 ?
    JP Z,scan_done

$$: CP.B (WP),R0  ; looking for 1st match...
    JP Z,scan_done
    INC WP
    DEC TOS
    JP NZ,$B
scan_done:
    PUSH WP  ; push c-addr', u' already in TOS
    next

; -----------------------------------------------------------
;Z S=    c-addr1 c-addr2 u -- n   string compare
;Z             n<0: s1<s2, n=0: s1=s2, n>0: s1>s2
    head SEQUAL,2,{"S="},docode

    ; count is in TOS
    OR TOS,TOS
    POP R1 ; string @ c-addr2
    POP R0 ; string @ c-addr1
    JP Z,smatch

$$: LD.UB WP,(R1++) ; char @ (c-addr2) -> WP
    CP.B (R0),WP    ; compare char @ (c-addr1) == WP 
    JP NZ,sdiff     ; outta here if mismatch
    INC R0          ; bump c-addr1
    DEC TOS         ; exhaust the count?
    JP NZ,$B        ; loop for length of string

smatch: ; count exhausted & no mismatch found
    CLR TOS     ; TOS=00000000  (s1=s2)
    JP $F

sdiff:
    ; mismatch!
    SBC TOS,TOS         ; propagate Carry-Flag
                        ; TOS=FFFFFFFF if cy (s1<s2)
    OR TOS,#1           ; TOS=00000001 if ncy (s1>s2)
$$:
    next

;Z BREAK   --  debugger break
    head BREAK,5,{"BREAK"},docode
    NOP ;use breakpoint here
    next

        .INCLUDE "CAML16Fd.S"   ; CPU Dependencies
        .INCLUDE "CAML16Fh.S"   ; High Level words
        .INCLUDE "CAML16Fu.S"   ; Utility/Other words

    XDEF lastword
lastword EQU dlink  ; nfa (name field address) of
                    ; last word in dictionary.

    XDEF InpBuffer
	SEGMENT NEAR_BSS
InpBuffer DS SIZE_TIB   ; space reserved for
                        ; for terminal input buffer

    XDEF enddict
enddict:
    ; user's code starts here
    SEGMENT FAR_BSS  ;Uninitialized far data
Caller_SP BLKL 1

    ;SEGMENT FAR_DATA ;Initialized far data
    END

